/* Copyright (C) 2010-2021 Frankfurt Institute for Advanced Studies, Goethe-Universität Frankfurt, Frankfurt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Igor Kulakov [committer], Maksym Zyzak, Valentina Akishina */

#include "L1Algo.h"

#include "CbmL1.h"

#include "CaGridEntry.h"
#include "CaToolsDebugger.h"

L1Algo::L1Algo()
{
  for (unsigned int i = 0; i < constants::size::MaxNstations; i++) {
    fTriplets[i].SetName(std::stringstream() << "L1Algo::fTriplets[" << i << "]");
  }
}

using cbm::algo::ca::EDetectorID;
using cbm::ca::tools::Debugger;

void L1Algo::Init(const TrackingMode mode) { fTrackingMode = mode; }

void L1Algo::Finish() { Debugger::Instance().Write(); }

// ---------------------------------------------------------------------------------------------------------------------
//
void L1Algo::ReceiveInputData(InputData&& inputData)
{
  // ----- Get input data ----------------------------------------------------------------------------------------------
  fInputData = std::move(inputData);
}



// ---------------------------------------------------------------------------------------------------------------------
//
void L1Algo::ReceiveParameters(Parameters&& parameters)
{
  fParameters = std::move(parameters);

  fNstationsBeforePipe = fParameters.GetNstationsActive(static_cast<EDetectorID>(0));

  // FIXME: SZh 24.08.2022
  //        This approach is suitable only for a case, when all the stations inside a magnetic field come right before
  //        all the stations outside of the field!
  fNfieldStations = std::lower_bound(fParameters.GetStations().cbegin(),
                                     fParameters.GetStations().cbegin() + fParameters.GetNstationsActive(),
                                     0,  // we are looking for the first zero element
                                     [](const ca::Station& s, int edge) { return bool(s.fieldStatus) > edge; })
                    - fParameters.GetStations().cbegin();

  fGhostSuppression = fParameters.GetGhostSuppression();

  ca::FieldRegion::ForceUseOfOriginalField(fParameters.DevIsUseOfOriginalField());
}

int L1Algo::GetMcTrackIdForCaHit(int iHit) const
{
  int hitId    = iHit;
  int iMcPoint = CbmL1::Instance()->GetHitBestMcRefs()[hitId];
  if (iMcPoint < 0) return -1;
  return CbmL1::Instance()->GetMcPoints()[iMcPoint].ID;
}

int L1Algo::GetMcTrackIdForWindowHit(int iHit) const
{
  int hitId    = fWindowHits[iHit].Id();
  int iMcPoint = CbmL1::Instance()->GetHitBestMcRefs()[hitId];
  if (iMcPoint < 0) return -1;
  return CbmL1::Instance()->GetMcPoints()[iMcPoint].ID;
}

const CbmL1MCTrack* L1Algo::GetMcTrackForWindowHit(int iHit) const
{
  int id = GetMcTrackIdForWindowHit(iHit);
  if (id < 0) return nullptr;
  return &CbmL1::Instance()->GetMcTracks()[id];
}

//   bool L1Algo::SortTrip(TripSort const& a, TripSort const& b) {
//       return   ( a.trip.GetLevel() >  b.trip.GetLevel() );
// }
//
// bool L1Algo::SortCand(CandSort const& a, CandSort const& b) {
//     if (a.cand.Lengtha != b.cand.Lengtha) return (a.cand.Lengtha > b.cand.Lengtha);
//
//     if (a.cand.ista != b.cand.ista ) return (a.cand.ista  < b.cand.ista );
//
//     if (a.cand.chi2  != b.cand.chi2 )return (a.cand.chi2  < b.cand.chi2 );
//     //return (a->chi2  < b->chi2 );
//     //   return (a->CandIndex < b->CandIndex );
//    // return (a.cand.CandIndex > b.cand.CandIndex );
// }

//   inline int L1Algo::PackIndex(const int& a, const int& b, const int& c) {
//       return   (a) + ((b)*10000) + (c*100000000);
// }
//
//   inline int L1Algo::UnPackIndex(const int& i, int& a, int& b, int& c) {
