/* Copyright (C) 2007-2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Maksym Zyzak, Valentina Akishina, Igor Kulakov [committer], Sergei Zharko */

#ifndef L1Algo_h
#define L1Algo_h

/// Debug features
// #define PULLS            // triplets pulls
// #define TRIP_PERFORMANCE // triplets efficiencies
// #define DOUB_PERFORMANCE // doublets efficiencies
// #define DRAW             // event display
#ifdef DRAW
class L1AlgoDraw;
#include "CbmL1Track.h"
#endif
//#define XXX               // time debug
//#define COUNTERS          // diff counters (hits, doublets, ... )

//#define GLOBAL
//#define mCBM

#define FIND_GAPED_TRACKS  // use triplets with gaps

#include <array>
#include <iomanip>
#include <iostream>
#include <limits>
#include <map>

#include "CaBranch.h"
#include "CaConstants.h"
#include "CaField.h"
#include "CaGrid.h"
#include "CaGridEntry.h"
#include "CaHit.h"
#include "CaInputData.h"
#include "CaParameters.h"
#include "CaStation.h"
#include "CaTrack.h"
#include "CaTrackParam.h"
#include "CaTriplet.h"
#include "CaVector.h"
#include "L1CloneMerger.h"
#include "L1Fit.h"
#include "L1Utils.h"

using namespace cbm::algo::ca;  //TODO: remove

class CbmL1MCTrack;

//namespace
//{
//  using cbm::algo::ca::Track;      // TMP
//  using cbm::algo::ca::Vector;     // TMP
//  using cbm::algo::ca::Iteration;  // TMP
//}  // namespace

// *******************************
// ** Types definition (global) **
// *******************************

using L1StationsArray_t = std::array<ca::Station, constants::size::MaxNstations>;
using L1MaterialArray_t = std::array<ca::MaterialMap, constants::size::MaxNstations>;
using Tindex            = int;  // TODO: Replace with ca::HitIndex_t, if suitable

#ifdef PULLS
#define TRIP_PERFORMANCE
class L1AlgoPulls;
#endif
#ifdef TRIP_PERFORMANCE
template<Tindex NHits>
class L1AlgoEfficiencyPerformance;
#endif
#ifdef DOUB_PERFORMANCE
template<Tindex NHits>
class L1AlgoEfficiencyPerformance;
#endif

struct L1HitTimeInfo {
  fscal fEventTimeMin {-std::numeric_limits<fscal>::max() / 2.};
  fscal fEventTimeMax {std::numeric_limits<fscal>::max() / 2.};
  fscal fMaxTimeBeforeHit {0.};  //< max event time for hits [0 .. hit] in the station hit array
  fscal fMinTimeAfterHit {0.};   //< min event time for hits [hit ... ] in the station hit array
};

/// Main class of L1 CA track finder algorithm
///
class L1Algo {
public:
  // *************************
  // ** Friend classes list **
  // *************************

  friend class CbmL1;  /// TODO: Remove this dependency
  friend class L1TripletConstructor;
#ifdef DRAW
  friend class L1AlgoDraw;
#endif

  // **********************************
  // ** Member functions declaration **
  // **********************************

  // ** Constructors and destructor

  /// Constructor
  L1Algo();

  /// Copy constructor
  L1Algo(const L1Algo&) = delete;

  /// Move constructor
  L1Algo(L1Algo&&) = delete;

  /// Copy assignment operator
  L1Algo& operator=(const L1Algo&) = delete;

  /// Move assignment operator
  L1Algo& operator=(L1Algo&&) = delete;

  /// Destructor
  ~L1Algo() = default;


  // ** Functions, which pack and unpack indexes of station and triplet **

  // TODO: move to ca::Triplet class (S.Zharko)
  /// Packs station and triplet indices to an unique triplet ID
  /// \param  iStation  Index of station in the active stations array
  /// \param  iTriplet  Index of triplet
  static unsigned int PackTripletId(unsigned int iStation, unsigned int iTriplet);

  /// Unpacks the triplet ID to its station index
  /// \param  id  Unique triplet ID
  static unsigned int TripletId2Station(unsigned int id);

  /// Unpacks the triplet ID to its triplet index
  /// \param  id  Unique triplet ID
  static unsigned int TripletId2Triplet(unsigned int id);


  /// Sets L1Algo parameters object
  /// \param other - reference to the Parameters object
  void SetParameters(const Parameters& other) { fParameters = other; }
  // TODO: remove it (S.Zharko)

  /// Gets a pointer to the L1Algo parameters object
  const Parameters* GetParameters() const { return &fParameters; }

  /// Receives input data
  void ReceiveInputData(InputData&& inputData);

  /// Receives tracking parameters
  void ReceiveParameters(Parameters&& parameters);

  /// Gets pointer to input data object for external access
  const InputData& GetInputData() const { return fInputData; }

  inline int PackIndex(const int& a, const int& b, const int& c);

  inline int UnPackIndex(const int& i, int& a, int& b, int& c);

  /// \brief Sets a default particle mass for the track fit
  /// It is used during reconstruction
  /// for the multiple scattering and energy loss estimation
  /// \param mass Default particle mass
  void SetDefaultParticleMass(float mass) { fDefaultMass = mass; }

  /// Gets default particle mass
  /// \return particle mass
  float GetDefaultParticleMass() const { return fDefaultMass; }

  /// Gets default particle mass squared
  /// \return particle mass squared
  float GetDefaultParticleMass2() const { return fDefaultMass * fDefaultMass; }


  /*********************************************************************************************/ /**
   *                             ------  FUNCTIONAL PART ------
   ************************************************************************************************/

  /// ----- Subroutines used by L1Algo::CaTrackFinder() ------

  void ReadWindowData();

  void CAFindTrack(int ista, ca::Branch& best_tr, const ca::Triplet* curr_trip, ca::Branch& curr_tr,
                   unsigned char min_best_l, ca::Branch* new_tr);


  /// Fits track
  /// \param t - track with hits
  /// \param T - track parameters
  /// \param dir - false - forward, true - backward
  /// \param qp0 - momentum for extrapolation
  /// \param initParams - should be params ititialized. 1 - yes.
  void BranchFitterFast(const ca::Branch& t, TrackParamV& T, const bool dir, const fvec qp0 = 0.,
                        const bool initParams = true);

  /// Fits track. more precise than FitterFast
  /// \param t - track with hits
  /// \param T - track parameters
  /// \param dir - false - forward, true - backward
  /// \param qp0 - momentum for extrapolation
  /// \param initParams - should be params ititialized. 1 - yes.
  void BranchFitter(const ca::Branch& t, TrackParamV& T, const bool dir, const fvec qp0 = 0.,
                    const bool initParams = true);

  /// Finds additional hits for already found track
  /// \param t - track with hits
  /// \param T - track params
  /// \param dir - 0 - forward, 1 - backward
  /// \param qp0 - momentum for extrapolation
  void FindMoreHits(ca::Branch& t, TrackParamV& T, const bool dir, const fvec qp0 = 0.0);

  /// Find additional hits for existing track
  /// \return chi2
  fscal BranchExtender(ca::Branch& t);

#ifdef DRAW
  L1AlgoDraw* draw {nullptr};
  void DrawRecoTracksTime(const Vector<CbmL1Track>& tracks);
#endif

  /// TODO: Move to L1
  enum TrackingMode
  {
    kSts,
    kGlobal,
    kMcbm
  };

  void Init(const TrackingMode mode);

  void Finish();

  void PrintHits();

  /// The main procedure - find tracks
  void CaTrackFinder();

  /// find tracks in a sub-timeslice
  void CaTrackFinderSlice();

  /// Track fitting procedures

  void L1KFTrackFitter();  // version from SIMD-KF benchmark


  float GetMaxInvMom() const { return fMaxInvMom[0]; }

public:
  /// Gets number of stations before the pipe (MVD stations in CBM)
  int GetNstationsBeforePipe() const { return fNstationsBeforePipe; }

  /// Gets number of stations situated in field region (MVD + STS in CBM)
  int GetNfieldStations() const { return fNfieldStations; }

  /// Get mc track ID for a hit (debug tool)
  int GetMcTrackIdForCaHit(int iHit) const;

  /// Get mc track ID for a hit (debug tool)
  int GetMcTrackIdForWindowHit(int iGridHit) const;

  const CbmL1MCTrack* GetMcTrackForWindowHit(int iHit) const;

private:
  bool checkTripletMatch(const ca::Triplet& l, const ca::Triplet& r, fscal& dchi2) const;

  int fNstationsBeforePipe {0};  ///< number of stations before pipe (MVD stations in CBM)
  int fNfieldStations {0};       ///< number of stations in the field region
  //alignas(16) L1StationsArray_t fStations {};  ///< array of ca::Station objects
  //alignas(16) L1MaterialArray_t fRadThick {};  ///< material for each station
  float fDefaultMass {constants::phys::MuonMass};  ///< mass of the propagated particle [GeV/c2]


  // ***************************
  // ** Member variables list **
  // ***************************

  Parameters fParameters;  ///< Object of L1Algo parameters class
  InputData fInputData;    ///< Tracking input data

  Vector<unsigned char> fvHitKeyFlags {
    "L1Algo::fvHitKeyFlags"};  ///< List of key flags: has been this hit or cluster already used

public:
  Vector<L1HitTimeInfo> fHitTimeInfo;

  ca::Grid vGrid[constants::size::MaxNstations];  ///<

  /// ----- Data used during track finding -----
  ///

  ///\brief hits of the current time window
  /// it is a portion of fInputData to process in the current time window
  /// hit.Id is replaced by the hit index in fInputData
  Vector<ca::Hit> fWindowHits {"L1Algo::fWindowHits"};

  Vector<unsigned char> fIsWindowHitSuppressed {"L1Algo::fIsWindowHitSuppressed"};

  ///\brief first index of the station hits in the time window
  ca::HitIndex_t fStationHitsStartIndex[constants::size::MaxNstations + 1] {0};

  ///\brief number of station hits in the time window
  ca::HitIndex_t fStationNhits[constants::size::MaxNstations + 1] {0};

  /// ----------

  double fCaRecoTime {0.};  // time of the track finder + fitter

  Vector<Track> fRecoTracks {"L1Algo::fRecoTracks"};       ///< reconstructed tracks
  Vector<ca::HitIndex_t> fRecoHits {"L1Algo::fRecoHits"};  ///< packed hits of reconstructed tracks

  Vector<Track> fSliceRecoTracks {"L1Algo::fSliceRecoTracks"};       ///< reconstructed tracks in sub-timeslice
  Vector<ca::HitIndex_t> fSliceRecoHits {"L1Algo::fSliceRecoHits"};  ///< packed hits of reconstructed tracks

  /// Created triplets vs station index
  Vector<ca::Triplet> fTriplets[constants::size::MaxNstations] {{"L1Algo::fTriplets"}};

  /// Track candidates created out of adjacent triplets before the final track selection.
  /// The candidates may share any amount of hits.
  Vector<ca::Branch> fTrackCandidates {"L1Algo::fTrackCandidates"};

  ///< indices of the sub-slice hits
  Vector<ca::HitIndex_t> fSliceHitIds[constants::size::MaxNstations] {"L1Algo::fSliceHitIds"};

  Vector<int> fStripToTrack {"L1Algo::fStripToTrack"};  // strip to track pointers

  TrackingMode fTrackingMode {kSts};

  fvec EventTime {0.f};
  fvec Err {0.f};

  /// --- data used during finding iterations
  unsigned int fCurrentIterationIndex {0};        ///< index of the corrent iteration, needed for debug output
  const Iteration* fpCurrentIteration = nullptr;  ///< pointer to the current CA track finder iteration

  Vector<int> fHitFirstTriplet {"L1Algo::fHitFirstTriplet"};  /// link hit -> first triplet { hit, *, *}
  Vector<int> fHitNtriplets {"L1Algo::fHitNtriplets"};        /// link hit ->n triplets { hit, *, *}


private:
  L1CloneMerger fCloneMerger {*this};  ///< Object of L1Algo clones merger algorithm

private:
  /// =================================  DATA PART  =================================

  /// ----- Different parameters of CATrackFinder -----

  Tindex fFirstCAstation {0};  //first station used in CA

  // fNFindIterations - set number of interation for trackfinding
  // itetation of finding:

  int fNFindIterations = -1;  // TODO investigate kAllPrimJumpIter & kAllSecJumpIter

  float fTrackChi2Cut {10.f};
  float fTripletFinalChi2Cut {10.f};
  float fTripletChi2Cut {5.f};  // cut for selecting triplets before collecting tracks.per one DoF
  float fDoubletChi2Cut {5.f};
  float fTimeCut1 {0.f};  // TODO: please, specify "1" and "2" (S.Zharko)
  float fTimeCut2 {0.f};

  /// correction in order to take into account overlaping and iff z. if sort by y then it is max diff between same station's modules (~0.4cm)
  fvec fMaxDZ {L1Utils::kNaN};

  /// parameters which are different for different iterations. Set in the begin of CAL1TrackFinder

  float fPickGather {L1Utils::kNaN};       ///< same for attaching additional hits to track
  float fTripletLinkChi2 {L1Utils::kNaN};  ///< (dp2/dp_error2 < fTripletLinkChi2)  =>  triplets are neighbours
  fvec fMaxInvMom {L1Utils::kNaN};         ///< max considered q/p for tracks
  fvec fMaxSlopePV {L1Utils::kNaN};        ///< max slope (tx\ty) in prim vertex
  float fMaxSlope {L1Utils::kNaN};         ///< max slope (tx\ty) in 3d hit position of a triplet

  fvec fTargX {L1Utils::kNaN};  ///< target position x coordinate for the current iteration (modifiable)
  fvec fTargY {L1Utils::kNaN};  ///< target position y coordinate for the current iteration (modifiable)
  fvec fTargZ {L1Utils::kNaN};  ///< target position z coordinate for the current iteration (modifiable)

  bool fIsTargetField {false};  ///< is the magnetic field present at the target

  ca::FieldValue fTargB _fvecalignment {};             // field in the target point (modifiable, do not touch!!)
  ca::MeasurementXy<fvec> TargetMeasurement _fvecalignment {};  // target constraint  [cm]

  int fGhostSuppression {1};  // NOTE: Should be equal to 0 in TRACKS_FROM_TRIPLETS mode!

  /// ----- Debug features -----
#ifdef PULLS
  L1AlgoPulls* fL1Pulls;
#endif
#ifdef TRIP_PERFORMANCE
  L1AlgoEfficiencyPerformance<3>* fL1Eff_triplets;
  L1AlgoEfficiencyPerformance<3>* fL1Eff_triplets2;
#endif
#ifdef DOUB_PERFORMANCE
  L1AlgoEfficiencyPerformance<2>* fL1Eff_doublets;
#endif
} _fvecalignment;


// ********************************************
// ** Inline member functions implementation **
// ********************************************

// ---------------------------------------------------------------------------------------------------------------------
//
[[gnu::always_inline]] inline unsigned int L1Algo::PackTripletId(unsigned int iStation, unsigned int iTriplet)
{
#ifndef FAST_CODE
  assert(iStation < constants::size::MaxNstations);
  assert(iTriplet < constants::size::MaxNtriplets);
#endif
  constexpr unsigned int kMoveStation = constants::size::TripletBits;
  return (iStation << kMoveStation) + iTriplet;
}

// ---------------------------------------------------------------------------------------------------------------------
//
[[gnu::always_inline]] inline unsigned int L1Algo::TripletId2Station(unsigned int id)
{
  constexpr unsigned int kMoveStation = constants::size::TripletBits;
  return id >> kMoveStation;
}

// ---------------------------------------------------------------------------------------------------------------------
//
[[gnu::always_inline]] inline unsigned int L1Algo::TripletId2Triplet(unsigned int id)
{
  constexpr unsigned int kTripletMask = (1u << constants::size::TripletBits) - 1u;
  return id & kTripletMask;
}

#endif
