/* Copyright (C) 2009-2021 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Valentina Akishina, Ivan Kisel, Sergey Gorbunov [committer], Igor Kulakov, Maksym Zyzak */

/*
 *=====================================================
 *
 *  CBM Level 1 4D Reconstruction
 *
 *  Authors: V.Akishina, I.Kisel,  S.Gorbunov, I.Kulakov, M.Zyzak
 *  Documentation: V.Akishina
 *
 *  e-mail : v.akishina@gsi.de
 *
 *=====================================================
 *
 *  Finds tracks using the Cellular Automaton algorithm
 *
 */


#include "CbmL1.h"
#include "CbmL1MCTrack.h"

#include "CaBranch.h"
#include "CaGrid.h"
#include "CaGridEntry.h"
#include "CaTrack.h"
#include "CaTrackParam.h"
#include "L1Algo.h"
#include "L1Assert.h"
#include "L1Fit.h"
#include "L1TripletConstructor.h"
#ifdef DRAW
#include "utils/L1AlgoDraw.h"
#endif  // DRAW
#ifdef PULLS
#include "utils/L1AlgoPulls.h"
#endif  // PULLS
#ifdef TRIP_PERFORMANCE
#include "utils/L1AlgoEfficiencyPerformance.h"
#endif  // TRIP_PERFORMANCE
#ifdef DOUB_PERFORMANCE
#include "utils/L1AlgoEfficiencyPerformance.h"
#endif  // DOUB_PERFORMANCE

#include <algorithm>
#include <fstream>
#include <iostream>
#include <list>
#include <map>

#include <stdio.h>

#include "CaToolsDebugger.h"


// using namespace std;
using std::cout;
using std::endl;
using Track = cbm::algo::ca::Track;

bool L1Algo::checkTripletMatch(const ca::Triplet& l, const ca::Triplet& r, fscal& dchi2) const
{
  dchi2 = 1.e20;

  if (r.GetMHit() != l.GetRHit()) return false;
  if (r.GetLHit() != l.GetMHit()) return false;

  if (r.GetMSta() != l.GetRSta()) return false;
  if (r.GetLSta() != l.GetMSta()) return false;


  if (r.IsMomentumFitted()) {
    assert(l.IsMomentumFitted());

    fscal dqp = l.GetQp() - r.GetQp();
    fscal Cqp = l.GetCqp() + r.GetCqp();

    if (!std::isfinite(dqp)) return false;
    if (!std::isfinite(Cqp)) return false;

    if (dqp * dqp > fTripletLinkChi2 * Cqp) {
      return false;  // bad neighbour // CHECKME why do we need recheck it?? (it really change result)
    }
    dchi2 = dqp * dqp / Cqp;
  }
  else {
    fscal dtx = l.GetTx() - r.GetTx();
    fscal Ctx = l.GetCtx() + r.GetCtx();

    fscal dty = l.GetTy() - r.GetTy();
    fscal Cty = l.GetCty() + r.GetCty();

    // it shouldn't happen, but happens sometimes

    if (!std::isfinite(dtx)) return false;
    if (!std::isfinite(dty)) return false;
    if (!std::isfinite(Ctx)) return false;
    if (!std::isfinite(Cty)) return false;

    if (dty * dty > fTripletLinkChi2 * Cty) return false;
    if (dtx * dtx > fTripletLinkChi2 * Ctx) return false;

    //dchi2 = 0.5f * (dtx * dtx / Ctx + dty * dty / Cty);
    dchi2 = 0.;
  }

  if (!std::isfinite(dchi2)) return false;

  return true;
}


// **************************************************************************************************
// *                                                                                                *
// *                            ------ CATrackFinder procedure ------                               *
// *                                                                                                *
// **************************************************************************************************

// ---------------------------------------------------------------------------------------------------------------------
//
void L1Algo::ReadWindowData()
{
  int nHits = 0;
  for (int iS = 0; iS < fParameters.GetNstationsActive(); iS++) {
    fStationHitsStartIndex[iS] = nHits;
    fStationNhits[iS]          = fSliceHitIds[iS].size();
    nHits += fStationNhits[iS];
  }
  fStationHitsStartIndex[fParameters.GetNstationsActive()] = nHits;

  fWindowHits.reset(nHits);
  fIsWindowHitSuppressed.reset(nHits, 0);

  for (int iS = 0; iS < fParameters.GetNstationsActive(); iS++) {
    for (ca::HitIndex_t ih = 0; ih < fSliceHitIds[iS].size(); ++ih) {
      ca::Hit h = fInputData.GetHit(fSliceHitIds[iS][ih]);
      h.SetId(fSliceHitIds[iS][ih]);

      fWindowHits[fStationHitsStartIndex[iS] + ih] = h;
    }
  }

  fHitFirstTriplet.reset(nHits, 0);
  fHitNtriplets.reset(nHits, 0);

  // TODO: SG: reduce the array size to the number of keys in subslice

  fStripToTrack.reset(fInputData.GetNhitKeys(), -1);

  fSliceRecoTracks.clear();
  fSliceRecoTracks.reserve(2 * nHits / fParameters.GetNstationsActive());

  fSliceRecoHits.clear();
  fSliceRecoHits.reserve(2 * nHits);

  fTrackCandidates.clear();
  fTrackCandidates.reserve(nHits / 10);
  for (unsigned int iS = 0; iS < constants::size::MaxNstations; iS++) {
    int nHitsStation = fSliceHitIds[iS].size();
    fTriplets[iS].clear();
    fTriplets[iS].reserve(2 * nHitsStation);
  }
}

void L1Algo::CaTrackFinderSlice()
{
  fNFindIterations = fParameters.GetNcaIterations();

  /*******************************/ /**
   * Performance monitors setting
   **********************************/

#ifdef PULLS
  static L1AlgoPulls* l1Pulls_ = new L1AlgoPulls();
  fL1Pulls                     = l1Pulls_;
  fL1Pulls->Init();
#endif
#ifdef TRIP_PERFORMANCE
  static L1AlgoEfficiencyPerformance<3>* l1Eff_triplets_ = new L1AlgoEfficiencyPerformance<3>();
  fL1Eff_triplets                                        = l1Eff_triplets_;
  fL1Eff_triplets->Init();
  static L1AlgoEfficiencyPerformance<3>* l1Eff_triplets2_ = new L1AlgoEfficiencyPerformance<3>();
  fL1Eff_triplets2                                        = l1Eff_triplets2_;
  fL1Eff_triplets2->Init();
#endif
#ifdef DOUB_PERFORMANCE
  static L1AlgoEfficiencyPerformance<2>* l1Eff_doublets_ = new L1AlgoEfficiencyPerformance<2>();
  fL1Eff_doublets                                        = l1Eff_doublets_;
  fL1Eff_doublets->Init();
#endif

#ifdef DRAW
  if (!draw) draw = new L1AlgoDraw;
  draw->InitL1Draw(this);
#endif


#if defined(COUNTERS)
  static unsigned int stat_N = 0;  // number of events
  stat_N++;

  static Tindex stat_nStartHits              = 0;
  static Tindex stat_nHits[fNFindIterations] = {0};

  static Tindex stat_nSinglets[fNFindIterations] = {0};
  //  static Tindex stat_nDoublets[fNFindIterations] = {0};
  static Tindex stat_nTriplets[fNFindIterations] = {0};

  static Tindex stat_nLevels[constants::size::MaxNstations - 2][fNFindIterations] = {{0}};
  static Tindex stat_nCalls[fNFindIterations]                                     = {0};  // n calls of CAFindTrack
  static Tindex stat_nTrCandidates[fNFindIterations]                              = {0};

  for (int iS = 0; iS < fParameters.GetNstationsActive(); ++iS) {
    stat_nStartHits += fSliceHitIds[iS].size();
  }
#endif

  /********************************/ /**
   * CATrackFinder routine setting
   ***********************************/

  ReadWindowData();

  for (int iS = 0; iS < fParameters.GetNstationsActive(); ++iS) {

    bool timeUninitialised = 1;
    fscal lasttime         = 0;
    fscal starttime        = 0;


    fscal gridMinX = -0.1;
    fscal gridMaxX = 0.1;
    fscal gridMinY = -0.1;
    fscal gridMaxY = 0.1;

    for (ca::HitIndex_t ih = 0; ih < fSliceHitIds[iS].size(); ++ih) {
      const ca::Hit& h = fInputData.GetHit(fSliceHitIds[iS][ih]);

      if (h.X() < gridMinX) { gridMinX = h.X(); }
      if (h.X() > gridMaxX) { gridMaxX = h.X(); }
      if (h.Y() < gridMinY) { gridMinY = h.Y(); }
      if (h.Y() > gridMaxY) { gridMaxY = h.Y(); }

      const fscal time = h.T();
      assert(std::isfinite(time));
      if (timeUninitialised || lasttime < time) { lasttime = time; }
      if (timeUninitialised || starttime > time) { starttime = time; }
      timeUninitialised = false;
    }

    // TODO: changing the grid also changes the result. Investigate why it happens.
    // TODO: find the optimal grid size

    int nSliceHits = fSliceHitIds[iS].size();
    fscal sizeY    = gridMaxY - gridMinY;
    fscal sizeX    = gridMaxX - gridMinX;
    int nBins2D    = 1 + nSliceHits;
    fscal yStep    = sizeY / sqrt(nBins2D);
    fscal xStep    = sizeX / sqrt(nBins2D);

    fscal scale = fParameters.GetStation(iS).GetZScal() - fParameters.GetTargetPositionZ()[0];

    if (yStep > 0.3 * scale) yStep = 0.3 * scale;
    if (yStep < 0.01 * scale) yStep = 0.01 * scale;
    if (xStep > 0.3 * scale) xStep = 0.3 * scale;
    if (xStep < 0.01 * scale) xStep = 0.01 * scale;

    vGrid[iS].BuildBins(gridMinX, gridMaxX, gridMinY, gridMaxY, xStep, yStep);
    vGrid[iS].StoreHits(fWindowHits, fStationHitsStartIndex[iS], fStationNhits[iS], fvHitKeyFlags);
  }


  // ---- Loop over Track Finder iterations ----------------------------------------------------------------//


  L1ASSERT(0, fNFindIterations == (int) fParameters.GetCAIterations().size());

  for (fCurrentIterationIndex = 0; fCurrentIterationIndex < fParameters.GetCAIterations().size();
       ++fCurrentIterationIndex) {

    // current iteration of the CA tracker
    const auto& caIteration = fParameters.GetCAIterations()[fCurrentIterationIndex];
    fpCurrentIteration      = &caIteration;

    if (fCurrentIterationIndex > 0) {
      for (int ista = 0; ista < fParameters.GetNstationsActive(); ++ista) {
        vGrid[ista].RemoveUsedHits(fWindowHits, fvHitKeyFlags);
      }
    }

    fIsWindowHitSuppressed.reset(fWindowHits.size(), 0);

    for (int j = 0; j < fParameters.GetNstationsActive(); j++) {
      fTriplets[j].clear();
    }

#ifdef COUNTERS
    Tindex nSinglets = 0;
#endif

    fHitFirstTriplet.reset(fWindowHits.size(), 0);
    fHitNtriplets.reset(fWindowHits.size(), 0);

    {
      // #pragma omp  task
      {
        // --- SET PARAMETERS FOR THE ITERATION ---

        fFirstCAstation      = caIteration.GetFirstStationIndex();
        fTrackChi2Cut        = caIteration.GetTrackChi2Cut();
        fDoubletChi2Cut      = caIteration.GetDoubletChi2Cut();  //11.3449 * 2.f / 3.f;  // prob = 0.1
        fTripletChi2Cut      = caIteration.GetTripletChi2Cut();  //21.1075;  // prob = 0.01%
        fTripletFinalChi2Cut = caIteration.GetTripletFinalChi2Cut();
        fPickGather          = caIteration.GetPickGather();       //3.0;
        fTripletLinkChi2     = caIteration.GetTripletLinkChi2();  //5.0;
        fMaxInvMom           = caIteration.GetMaxQp();            //1.0 / 0.5;  // max considered q/p
        fMaxSlopePV          = caIteration.GetMaxSlopePV();       //1.1;
        fMaxSlope            = caIteration.GetMaxSlope();         //2.748;  // corresponds to 70 grad

        // define the target
        fTargX = fParameters.GetTargetPositionX();
        fTargY = fParameters.GetTargetPositionY();
        fTargZ = fParameters.GetTargetPositionZ();

        fscal SigmaTargetX = caIteration.GetTargetPosSigmaX();
        fscal SigmaTargetY = caIteration.GetTargetPosSigmaY();  // target constraint [cm]

        // Select magnetic field. For primary tracks - fVtxFieldValue, for secondary tracks - st.fieldSlice
        if (caIteration.GetPrimaryFlag()) { fTargB = fParameters.GetVertexFieldValue(); }
        else {
          fTargB = fParameters.GetStation(0).fieldSlice.GetFieldValue(0, 0);
        }  // NOTE: calculates field fTargB in the center of 0th station

        fIsTargetField = (fabs(fTargB.y[0]) > 0.001);

        TargetMeasurement.SetX(fTargX);
        TargetMeasurement.SetY(fTargY);
        TargetMeasurement.SetDx2(SigmaTargetX * SigmaTargetX);
        TargetMeasurement.SetDxy(0);
        TargetMeasurement.SetDy2(SigmaTargetY * SigmaTargetY);
        TargetMeasurement.SetNdfX(1);
        TargetMeasurement.SetNdfY(1);

        /// Set correction in order to take into account overlaping and iff z.
        /// The reason is that low momentum tracks are too curved and goes not from target direction. That's why sort by hit_y/hit_z is not work idealy
        /// If sort by y then it is max diff between same station's modules (~0.4cm)
        fMaxDZ = caIteration.GetMaxDZ();  //0;
      }
    }


    ///   stage for triplets creation

    L1TripletConstructor constructor1(this);
    L1TripletConstructor constructor2(this);
    L1TripletConstructor constructor3(this);

    for (int istal = fParameters.GetNstationsActive() - 2; istal >= fFirstCAstation;
         istal--) {  // start with downstream chambers
      Tindex nGridEntriesL = vGrid[istal].GetEntries().size();

      for (Tindex iel = 0; iel < nGridEntriesL; ++iel) {
        ca::HitIndex_t ihitl = vGrid[istal].GetEntries()[iel].GetObjectId();
        constructor1.CreateTripletsForHit(istal, istal + 1, istal + 2, ihitl);

        if (fpCurrentIteration->GetJumpedFlag()) {
          constructor2.CreateTripletsForHit(istal, istal + 1, istal + 3, ihitl);
          constructor3.CreateTripletsForHit(istal, istal + 2, istal + 3, ihitl);
        }

        auto oldSize = fTriplets[istal].size();

        fTriplets[istal].insert(fTriplets[istal].end(), constructor1.GetTriplets().begin(),
                                constructor1.GetTriplets().end());
        fTriplets[istal].insert(fTriplets[istal].end(), constructor2.GetTriplets().begin(),
                                constructor2.GetTriplets().end());
        fTriplets[istal].insert(fTriplets[istal].end(), constructor3.GetTriplets().begin(),
                                constructor3.GetTriplets().end());

        fHitFirstTriplet[ihitl] = PackTripletId(istal, oldSize);
        fHitNtriplets[ihitl]    = fTriplets[istal].size() - oldSize;
      }
    }  // istal


    // search for neighbouring triplets

    for (int istal = fParameters.GetNstationsActive() - 2; istal >= fFirstCAstation;
         istal--) {  // start with downstream chambers

      for (unsigned int it = 0; it < fTriplets[istal].size(); ++it) {

        ca::Triplet& tr = fTriplets[istal][it];
        tr.SetLevel(0);
        tr.SetFNeighbour(0);
        tr.SetNNeighbours(0);

        if (istal > (fParameters.GetNstationsActive() - 4)) break;

        unsigned int nNeighbours = fHitNtriplets[tr.GetMHit()];

        unsigned int neighLocation = fHitFirstTriplet[tr.GetMHit()];
        unsigned int neighStation  = TripletId2Station(neighLocation);
        unsigned int neighTriplet  = TripletId2Triplet(neighLocation);

        if (nNeighbours > 0) { assert((int) neighStation == istal + 1 || (int) neighStation == istal + 2); }

        unsigned char level = 0;

        for (unsigned int iN = 0; iN < nNeighbours; ++iN, ++neighTriplet, ++neighLocation) {

          ca::Triplet& neighbour = fTriplets[neighStation][neighTriplet];

          fscal dchi2 = 0.;
          if (!checkTripletMatch(tr, neighbour, dchi2)) continue;

          if (tr.GetFNeighbour() == 0) { tr.SetFNeighbour(neighLocation); }

          tr.SetNNeighbours(neighLocation - tr.GetFNeighbour() + 1);

          const unsigned char neighbourLevel = neighbour.GetLevel();

          if (level <= neighbourLevel) { level = neighbourLevel + 1; }
        }
        tr.SetLevel(level);
      }  // neighbour search
    }    // istal

    ///====================================================================
    ///=                                                                  =
    ///=        Collect track candidates. CREATE TRACKS                   =
    ///=                                                                  =
    ///====================================================================

    // min level to start triplet. So min track length = min_level+3.
    int min_level = std::min(caIteration.GetMinNhits(), caIteration.GetMinNhitsStation0()) - 3;

    // TODO: SZh 04.10.2022: Why this fatal error is called?
    // NOTE: This line was wrapped into TRACKS_FROM_TRIPLETS ifdef
    //LOG(FATAL) << "L1CATrackFinder: min_level undefined in " << __FILE__ << " : " << __LINE__;
    if (fpCurrentIteration->GetTrackFromTripletsFlag()) { min_level = 0; }

    ca::Branch curr_tr;
    ca::Branch new_tr[constants::size::MaxNstations];
    ca::Branch best_tr;

    int ndf = 1;


    // collect consequtive: the longest tracks, shorter, more shorter and so on
    for (int firstTripletLevel = fParameters.GetNstationsActive() - 3; firstTripletLevel >= min_level;
         firstTripletLevel--) {
      // choose length in triplets number - firstTripletLevel - the maximum possible triplet level among all triplets in the searched candidate

      //  how many levels to check
      int nlevel = (fParameters.GetNstationsActive() - 2) - firstTripletLevel + 1;

      const unsigned char min_best_l =
        (firstTripletLevel > min_level) ? firstTripletLevel + 2 : min_level + 3;  // loose maximum


      fTrackCandidates.clear();

      fStripToTrack.reset(fInputData.GetNhitKeys(), -1);

      //== Loop over triplets with the required level, find and store track candidates

      for (int istaF = fFirstCAstation; istaF <= fParameters.GetNstationsActive() - 3 - firstTripletLevel; ++istaF) {

        if (--nlevel == 0) break;  //TODO: SG: this is not needed

        for (Tindex itrip = 0; itrip < (Tindex) fTriplets[istaF].size(); ++itrip) {

          ca::Triplet& first_trip = (fTriplets[istaF][itrip]);
          if (fvHitKeyFlags[fWindowHits[first_trip.GetLHit()].FrontKey()]
              || fvHitKeyFlags[fWindowHits[first_trip.GetLHit()].BackKey()]) {
            continue;
          }
          //               ghost suppression !!!
          //

          if (!fpCurrentIteration->GetTrackFromTripletsFlag()) {  // ghost suppression !!!
            int nHits = 3 + first_trip.GetLevel();
            if (fWindowHits[first_trip.GetLHit()].Station() == 0) {
              if (nHits < fpCurrentIteration->GetMinNhitsStation0()) { continue; }
            }
            else {
              if (nHits < fpCurrentIteration->GetMinNhits()) { continue; }
            }
          }

          // Collect triplets, which can start a track with length equal to firstTipletLevel + 3. This cut suppresses
          // ghost tracks, but does not affect the efficiency
          if (first_trip.GetLevel() < firstTripletLevel) { continue; }

          curr_tr.SetChi2(0.);
          curr_tr.SetStation(0);
          curr_tr.ResetHits();
          curr_tr.AddHit(fWindowHits[first_trip.GetLHit()].Id());
          curr_tr.SetChi2(first_trip.GetChi2());

          best_tr = curr_tr;

          CAFindTrack(istaF, best_tr, &first_trip, curr_tr, min_best_l,
                      new_tr);  /// reqursive func to build a tree of possible track-candidates and choose the best

          if (best_tr.NofHits() < firstTripletLevel + 2) continue;  // loose maximum one hit

          if (best_tr.NofHits() < min_level + 3) continue;  // should find all hits for min_level

          ndf = best_tr.NofHits() * 2 - 5;

          // TODO: automatize the NDF calculation

          if (kGlobal == fTrackingMode || kMcbm == fTrackingMode) { ndf = best_tr.NofHits() * 2 - 4; }

          best_tr.SetChi2(best_tr.Chi2() / ndf);
          if (fGhostSuppression) {
            if (3 == best_tr.NofHits()) {
              if (!fpCurrentIteration->GetPrimaryFlag() && (istaF != 0)) continue;  // too /*short*/ non-MAPS track
              if (fpCurrentIteration->GetPrimaryFlag() && (best_tr.Chi2() > 5.0)) continue;
            }
          }
          fTrackCandidates.push_back(best_tr);
          ca::Branch& tr = fTrackCandidates.back();
          tr.SetStation(istaF);
          tr.SetId(fTrackCandidates.size() - 1);
          tr.SetAlive(true);
          if (0) {  // debug
            cout << "track " << fTrackCandidates.size() - 1 << " found, L = " << best_tr.NofHits()
                 << " chi2= " << best_tr.Chi2() << endl;
            cout << " hits: ";
            for (auto hitId : tr.Hits()) {
              cout << GetMcTrackIdForCaHit(hitId) << " ";
            }
            cout << endl;
          }
        }  // itrip
      }    // istaF

      for (ca::Branch& tr : fTrackCandidates) {
        tr.SetAlive(false);
      }

      bool repeatCompetition = true;
      int iComp              = 0;
      for (iComp = 0; (iComp < 100) && repeatCompetition; ++iComp) {

        // == Loop over track candidates and mark their strips

        repeatCompetition = false;

        for (ca::Branch& tr : fTrackCandidates) {
          if (tr.IsAlive()) { continue; }
          for (auto& hitId : tr.Hits()) {
            const ca::Hit& h = fInputData.GetHit(hitId);
            bool isAlive     = true;
            {  // front  strip
              auto& stripF = (fStripToTrack)[h.FrontKey()];
              if ((stripF >= 0) && (stripF != tr.Id())) {  // strip is used by other candidate
                const auto& other = fTrackCandidates[stripF];
                if (!other.IsAlive() && tr.IsBetterThan(other)) { stripF = tr.Id(); }
                else {
                  isAlive = false;
                }
              }
              else {
                stripF = tr.Id();
              }
              if (!isAlive) { break; }
            }

            {  // back strip
              auto& stripB = (fStripToTrack)[h.BackKey()];
              if ((stripB >= 0) && (stripB != tr.Id())) {  // strip is used by other candidate
                const auto& other = fTrackCandidates[stripB];
                if (!other.IsAlive() && tr.IsBetterThan(other)) { stripB = tr.Id(); }
                else {
                  isAlive = false;
                }
              }
              else {
                stripB = tr.Id();
              }
              if (!isAlive) { break; }
            }
          }  // loop over hits
        }    // itrack

        // == Check if some suppressed candidates are still alive

        for (ca::Branch& tr : fTrackCandidates) {
          if (tr.IsAlive()) { continue; }

          tr.SetAlive(true);
          for (int iHit = 0; tr.IsAlive() && (iHit < (int) tr.Hits().size()); ++iHit) {
            const ca::Hit& h = fInputData.GetHit(tr.Hits()[iHit]);
            tr.SetAlive(tr.IsAlive() && ((fStripToTrack)[h.FrontKey()] == tr.Id())
                        && ((fStripToTrack)[h.BackKey()] == tr.Id()));
          }

          if (!tr.IsAlive()) {  // release strips
            for (auto hitId : tr.Hits()) {
              const ca::Hit& h = fInputData.GetHit(hitId);
              if (fStripToTrack[h.FrontKey()] == tr.Id()) { fStripToTrack[h.FrontKey()] = -1; }
              if (fStripToTrack[h.BackKey()] == tr.Id()) { fStripToTrack[h.BackKey()] = -1; }
            }
          }
          else {
            repeatCompetition = true;
          }
        }  // itrack

      }  // competitions

      // cout << " N Competitions: " << iComp << endl;

      // ==

      for (Tindex iCandidate = 0; iCandidate < (Tindex) fTrackCandidates.size(); ++iCandidate) {
        ca::Branch& tr = fTrackCandidates[iCandidate];
        if (!tr.IsAlive()) continue;
        if (kMcbm == fTrackingMode) {
          if (tr.NofHits() <= 3) { continue; }
        }
        else if (kGlobal == fTrackingMode) {
          if (tr.NofHits() < 3) { continue; }
        }
        else {
          if (tr.NofHits() < 3) { continue; }
        }
        if (fpCurrentIteration->GetExtendTracksFlag()) {
          if (tr.NofHits() < fParameters.GetNstationsActive()) { BranchExtender(tr); }
        }

        for (auto iHit : tr.Hits()) {
          const ca::Hit& hit = fInputData.GetHit(iHit);

          /// used strips are marked

          fvHitKeyFlags[hit.FrontKey()] = 1;
          fvHitKeyFlags[hit.BackKey()]  = 1;

          fSliceRecoHits.push_back(iHit);
        }
        Track t;
        t.fNofHits = tr.NofHits();
        fSliceRecoTracks.push_back(t);
        if (0) {  // SG debug
          cout << "store track " << iCandidate << " chi2= " << tr.Chi2() << endl;
          cout << " hits: ";
          for (unsigned int ih = 0; ih < tr.Hits().size(); ih++) {
            cout << GetMcTrackIdForCaHit(tr.Hits()[ih]) << " ";
          }
          cout << '\n';
        }

      }  // tracks

    }  // firstTripletLevel


    // suppress strips of suppressed hits
    for (unsigned int ih = 0; ih < fWindowHits.size(); ih++) {
      if (fIsWindowHitSuppressed[ih]) {
        const ca::Hit& hit            = fWindowHits[ih];
        fvHitKeyFlags[hit.FrontKey()] = 1;
        fvHitKeyFlags[hit.BackKey()]  = 1;
      }
    }

    {
      // #ifdef DRAW
      //     draw->ClearVeiw();
      // //   draw->DrawInfo();
      //     draw->DrawRestHits(fGridHitStartIndex, fGridHitStopIndex, RealIHit);
      //     draw->DrawRecoTracks();
      //     draw->SaveCanvas("Reco_"+iCaIteration+"_");
      //     draw->DrawAsk();
      // #endif

      // #ifdef PULLS
      //       fL1Pulls->Build(1);
      // #endif
#ifdef COUNTERS
      //  stat_nHits[iCaIteration] += (fGridHits*)->Size();
      cout << "iter = " << iCaIteration << endl;
      cout << " NSinglets = " << stat_nSinglets[iCaIteration] / stat_N << endl;
      //    cout << " NDoublets = " << stat_nDoublets[iCaIteration]/stat_N << endl;
      cout << " NTriplets = " << stat_nTriplets[iCaIteration] / stat_N << endl;
      cout << " NfGridHit = " << stat_nHits[iCaIteration] / stat_N << endl;
#endif  // COUNTERS
    }

  }  // ---- Loop over Track Finder iterations: END -----------------------------------------------------------//

  // Fit tracks
  this->L1KFTrackFitter();

  // Merge clones
  fCloneMerger.Exec(fSliceRecoTracks, fSliceRecoHits);

  //==================================

#ifdef DRAW
  draw->ClearVeiw();
  //   draw->DrawInputHits();
  //   draw->DrawInfo();
  draw->DrawRecoTracks();

  draw->SaveCanvas("Reco_");
  draw->DrawAsk();
#endif
#ifdef PULLS
  static int iEvee = 0;
  iEvee++;
  if (iEvee % 1 == 0) fL1Pulls->Build(1);
#endif
#ifdef DOUB_PERFORMANCE
  fL1Eff_doublets->CalculateEff();
  fL1Eff_doublets->Print("Doublets performance.", 1);
#endif
#ifdef TRIP_PERFORMANCE
  fL1Eff_triplets->CalculateEff();
  fL1Eff_triplets->Print("Triplet performance", 1);
//   fL1Eff_triplets2->CalculateEff();
//   fL1Eff_triplets2->Print("Triplet performance. After chi2 cut");
#endif
}


/** *************************************************************
     *                                                              *
     *     The routine performs recursive search for tracks         *
     *                                                              *
     *     I. Kisel                                    06.03.05     *
     *     I.Kulakov                                    2012        *
     *                                                              *
     ****************************************************************/

void L1Algo::CAFindTrack(int ista, ca::Branch& best_tr, const ca::Triplet* curr_trip, ca::Branch& curr_tr,
                         unsigned char min_best_l, ca::Branch* new_tr)
/// recursive search for tracks
/// input: @ista - station index, @&best_tr - best track for the privious call
/// output: @&NCalls - number of function calls
{

  if (curr_trip->GetLevel() == 0)  // the end of the track -> check and store
  {
    // -- finish with current track
    // add rest of hits
    const ca::HitIndex_t& ihitm = curr_trip->GetMHit();
    const ca::HitIndex_t& ihitr = curr_trip->GetRHit();


    //if (!GetFUsed((*fStripFlag)[fGridHits[ihitm].f] | (*fStripFlag)[fGridHits[ihitm].b])) {
    //L1_SHOW(fInputData.GetNhitKeys());
    //L1_SHOW(fvHitKeyFlags.size());
    //L1_SHOW(fGridHits[ihitm].f);
    //L1_SHOW(fGridHits[ihitm].b);
    if (!(fvHitKeyFlags[fWindowHits[ihitm].FrontKey()] || fvHitKeyFlags[fWindowHits[ihitm].BackKey()])) {

      //        curr_tr.Hits.push_back(fGridHitIds[ihitm]);

      curr_tr.AddHit(fWindowHits[ihitm].Id());
    }

    //if (!GetFUsed((*fStripFlag)[fGridHits[ihitr].f] | (*fStripFlag)[fGridHits[ihitr].b])) {
    //L1_SHOW(fInputData.GetNhitKeys());
    //L1_SHOW(fvHitKeyFlags.size());
    //L1_SHOW(fGridHits[ihitr].f);
    //L1_SHOW(fGridHits[ihitr].b);
    if (!(fvHitKeyFlags[fWindowHits[ihitr].FrontKey()] || fvHitKeyFlags[fWindowHits[ihitr].BackKey()])) {

      //curr_tr.Hits.push_back(fGridHitIds[ihitr]);
      curr_tr.AddHit(fWindowHits[ihitr].Id());
    }

    //if( curr_tr.NofHits() < min_best_l - 1 ) return; // suppouse that only one hit can be added by extender
    // TODO: SZh 21.08.2023: Replace hard-coded value with a parameter
    int ndf = curr_tr.NofHits() * 2 - 5;

    if (kGlobal == fTrackingMode || kMcbm == fTrackingMode) { ndf = curr_tr.NofHits() * 2 - 4; }

    if (curr_tr.Chi2() > fTrackChi2Cut * ndf) { return; }


    // -- select the best
    if ((curr_tr.NofHits() > best_tr.NofHits())
        || ((curr_tr.NofHits() == best_tr.NofHits()) && (curr_tr.Chi2() < best_tr.Chi2()))) {
      best_tr = curr_tr;
    }

    return;
  }
  else  //MEANS level ! = 0
  {
    int N_neighbour = (curr_trip->GetNNeighbours());

    for (Tindex in = 0; in < N_neighbour; in++) {

      unsigned int ID = curr_trip->GetFNeighbour() + in;

      unsigned int Station = TripletId2Station(ID);
      unsigned int Triplet = TripletId2Triplet(ID);

      const ca::Triplet& new_trip = fTriplets[Station][Triplet];

      fscal dchi2 = 0.;
      if (!checkTripletMatch(*curr_trip, new_trip, dchi2)) continue;

      if (fvHitKeyFlags[fWindowHits[new_trip.GetLHit()].FrontKey()]
          || fvHitKeyFlags[fWindowHits[new_trip.GetLHit()].BackKey()]) {  //hits are used
        //  no used hits allowed -> compare and store track
        if ((curr_tr.NofHits() > best_tr.NofHits())
            || ((curr_tr.NofHits() == best_tr.NofHits()) && (curr_tr.Chi2() < best_tr.Chi2()))) {
          best_tr = curr_tr;
        }
      }
      else {  // hit is not used: add the left hit from the new triplet to the current track

        unsigned char new_L = curr_tr.NofHits() + 1;
        fscal new_chi2      = curr_tr.Chi2() + dchi2;


        if (0) {  //SGtrd2d debug!!
          int mc01 = GetMcTrackIdForWindowHit(curr_trip->GetLHit());
          int mc02 = GetMcTrackIdForWindowHit(curr_trip->GetMHit());
          int mc03 = GetMcTrackIdForWindowHit(curr_trip->GetRHit());
          int mc11 = GetMcTrackIdForWindowHit(new_trip.GetLHit());
          int mc12 = GetMcTrackIdForWindowHit(new_trip.GetMHit());
          int mc13 = GetMcTrackIdForWindowHit(new_trip.GetRHit());
          if ((mc01 == mc02) && (mc02 == mc03)) {
            cout << " sta " << ista << " mc0 " << mc01 << " " << mc02 << " " << mc03 << " mc1 " << mc11 << " " << mc12
                 << " " << mc13 << " chi2 " << curr_tr.Chi2() / (2 * (curr_tr.NofHits() + 2) - 4) << " new "
                 << new_chi2 / (2 * (new_L + 2) - 4) << endl;
            cout << "   hits " << curr_trip->GetLHit() << " " << curr_trip->GetMHit() << " " << curr_trip->GetRHit()
                 << " " << new_trip.GetLHit() << endl;
          }
        }

        int ndf = 2 * (new_L + 2) - 5;

        if (kGlobal == fTrackingMode) {  //SGtrd2d!!!
          ndf = 2 * (new_L + 2) - 4;
        }
        else if (kMcbm == fTrackingMode) {
          ndf = 2 * (new_L + 2) - 4;
        }
        else {
          ndf = new_L;  // TODO: SG:  2 * (new_L + 2) - 5
        }

        if (new_chi2 > fTrackChi2Cut * ndf) continue;

        // add new hit
        new_tr[ista] = curr_tr;
        new_tr[ista].AddHit(fWindowHits[new_trip.GetLHit()].Id());

        const int new_ista = ista + new_trip.GetMSta() - new_trip.GetLSta();
        new_tr[ista].SetChi2(new_chi2);

        CAFindTrack(new_ista, best_tr, &new_trip, new_tr[ista], min_best_l, new_tr);
      }  // add triplet to track
    }    // for neighbours
  }      // level = 0
}

#ifdef DRAW
void L1Algo::DrawRecoTracksTime(const cbm::algo::ca::Vector<CbmL1Track>& tracks)
{
  // TODO: find where the DrawRecoTracksTime is. It is missing in the git repository.
  //draw->DrawRecoTracksTime(tracks);
  draw->SaveCanvas(" ");
}
#endif
