/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov [committer] */

#ifndef L1TripletConstructor_h
#define L1TripletConstructor_h

#include "CaField.h"
#include "CaGridEntry.h"
#include "CaStation.h"
#include "CaTrackParam.h"
#include "CaTriplet.h"
#include "CaVector.h"
#include "L1Algo.h"
#include "L1Fit.h"

namespace
{  // TMP!!
  using cbm::algo::ca::Vector;
}


/// Construction of triplets for the CA tracker
///
class L1TripletConstructor {
public:
  ///                             ------  Constructors and destructor ------

  /// Constructor
  /// \param nThreads  Number of threads for multi-threaded mode
  L1TripletConstructor(L1Algo* algo);

  /// Copy constructor
  L1TripletConstructor(const L1TripletConstructor&) = delete;

  /// Move constructor
  L1TripletConstructor(L1TripletConstructor&&) = delete;

  /// Copy assignment operator
  L1TripletConstructor& operator=(const L1TripletConstructor&) = delete;

  /// Move assignment operator
  L1TripletConstructor& operator=(L1TripletConstructor&&) = delete;

  /// Destructor
  ~L1TripletConstructor() = default;


  ///                             ------  FUNCTIONAL PART ------

  void InitStations(int istal, int istam, int istar);

  void CreateTripletsForHit(int istal, int istam, int istar, ca::HitIndex_t ihl);

  const Vector<ca::Triplet>& GetTriplets() const { return fTriplets; }

  /// Find the doublets. Reformat data in the portion of doublets.
  void FindDoublets();
  void FitDoublets();

  /// Find triplets on station
  void FindTriplets();

  /// Add the middle hits to parameters estimation. Propagate to right station.
  /// Find the triplets (right hit). Reformat data in the portion of triplets.
  void FindRightHit();

  /// Fit Triplets
  void FitTriplets();

  /// Select triplets. Save them into vTriplets.
  void StoreTriplets();

  void CollectHits(const TrackParamV& Tr, const int iSta, const double chi2Cut, const int iMC,
                   Vector<ca::HitIndex_t>& collectedHits, int maxNhits);

private:
  /// left station
  const ca::Station& staL() { return *fStaL; }
  /// middle station
  const ca::Station& staM() { return *fStaM; }
  /// right station
  const ca::Station& staR() { return *fStaR; }

private:
  bool fIsInitialized {false};  ///< is initialized;
  L1Algo* fAlgo {nullptr};      ///< pointer to L1Algo object

  int fIstaL {-1};  ///< left station index
  int fIstaM {-1};  ///< middle station index
  int fIstaR {-1};  ///< right station index

  const ca::Station* fStaL {nullptr};  ///< left station
  const ca::Station* fStaM {nullptr};  ///< mid station
  const ca::Station* fStaR {nullptr};  ///< right station

  const ca::Station* fFld0Sta[2];  // two stations for approximating the field between the target and the left hit
  const ca::Station* fFld1Sta[3];  // three stations for approximating the field between the left and the right hit

  ca::HitIndex_t fIhitL;  ///< index of the left hit in fAlgo->fWindowHits
  TrackParamV fTrL;
  ca::FieldRegion fFldL;

  Vector<ca::HitIndex_t> fHitsM_2 {"L1TripletConstructor::fHitsM_2"};
  Vector<TrackParamV> fTracks_2 {"L1TripletConstructor::fTracks_2"};

  Vector<TrackParamV> fTracks_3 {"L1TripletConstructor::fTracks_3"};
  Vector<ca::HitIndex_t> fHitsM_3 {"L1TripletConstructor::fHitsM_3"};
  Vector<ca::HitIndex_t> fHitsR_3 {"L1TripletConstructor::fHitsR_3"};

  Vector<ca::Triplet> fTriplets {"L1TripletConstructor::fTriplets"};

  L1Fit fFit;

} _fvecalignment;

#endif
